from __future__ import annotations

import itertools
from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import Iterable, Iterator
from dataclasses import dataclass, field
from typing import cast

from cognite.client.data_classes import data_modeling as dm
from cognite.client.data_classes.data_modeling.data_types import ListablePropertyType
from typing_extensions import Self

from cognite.pygen import config as pygen_config
from cognite.pygen.utils.text import create_name

_PRIMITIVE_TYPES = (dm.Text, dm.Boolean, dm.Float32, dm.Float64, dm.Int32, dm.Int64, dm.Timestamp, dm.Date, dm.Json)
_EXTERNAL_TYPES = (dm.TimeSeriesReference, dm.FileReference, dm.SequenceReference)


@dataclass(frozen=True)
class ViewSpaceExternalId:
    """
    This represents a view id.

    The motivation for this class is that API class and data classes are independent of the view version (given
    that all properties are equal). This enables reuse of data classes and APIs across view versions.

    Note this could be considered an edge case not worth supporting. However, when you publish data models
    from a .graphql format, the autogenerated view versions for identical types will be different when those types
    are part of two different data models.
    """

    space: str
    external_id: str

    @classmethod
    def from_(cls, view_id: dm.ViewId | dm.View) -> Self:
        return cls(space=view_id.space, external_id=view_id.external_id)


@dataclass(frozen=True)
class Field(ABC):
    """
    A field represents a pydantic field in the generated pydantic class.

    Args:
        name: The name of the field. This is used in the generated Python code.
        prop_name: The name of the property in the data model. This is used when reading and writing to CDF.
        pydantic_field: The name to use for the import 'from pydantic import Field'. This is used in the edge case
                        when the name 'Field' name clashes with the data model class name.

    """

    name: str
    prop_name: str
    pydantic_field: str

    @property
    def need_alias(self) -> bool:
        return self.name != self.prop_name

    @classmethod
    def from_property(
        cls,
        prop_name: str,
        prop: dm.MappedProperty | dm.ConnectionDefinition,
        data_class_by_view_id: dict[ViewSpaceExternalId, DataClass],
        field_naming: pygen_config.FieldNaming,
        view_name: str,
        pydantic_field: str = "Field",
    ) -> Field:
        name = create_name(prop_name, field_naming.name)
        if isinstance(prop, dm.SingleHopConnectionDefinition):
            variable = create_name(prop_name, field_naming.variable)

            edge_api_class_input = f"{view_name}_{prop_name}"
            edge_api_class = f"{create_name(edge_api_class_input, field_naming.edge_api_class)}API"
            edge_api_attribute = create_name(prop_name, field_naming.api_class_attribute)
            return EdgeOneToMany(
                name=name,
                prop_name=prop_name,
                prop=prop,
                data_class=data_class_by_view_id[ViewSpaceExternalId(prop.source.space, prop.source.external_id)],
                variable=variable,
                pydantic_field=pydantic_field,
                edge_api_class=edge_api_class,
                edge_api_attribute=edge_api_attribute,
            )
        elif isinstance(prop, dm.MappedProperty) and (
            isinstance(prop.type, _PRIMITIVE_TYPES) or isinstance(prop.type, _EXTERNAL_TYPES)
        ):
            type_ = _to_python_type(prop.type)
            if isinstance(prop.type, ListablePropertyType) and prop.type.is_list:
                return PrimitiveListField(
                    name=name,
                    prop_name=prop_name,
                    type_=type_,
                    is_nullable=prop.nullable,
                    prop=prop,
                    pydantic_field=pydantic_field,
                )
            elif isinstance(prop.type, dm.CDFExternalIdReference):
                # Note: these are only CDF External Fields that are not listable. Listable CDF External Fields
                # are handled above.
                edge_api_class_input = f"{view_name}_{prop_name}"
                edge_api_class = f"{create_name(edge_api_class_input, field_naming.edge_api_class)}"
                edge_api_attribute = create_name(prop_name, field_naming.api_class_attribute)
                return CDFExternalField(
                    name=name,
                    prop_name=prop_name,
                    type_=type_,
                    is_nullable=prop.nullable,
                    prop=prop,
                    pydantic_field=pydantic_field,
                    edge_api_class=edge_api_class,
                    edge_api_attribute=edge_api_attribute,
                )
            else:
                return PrimitiveField(
                    name=name,
                    prop_name=prop_name,
                    type_=type_,
                    is_nullable=prop.nullable,
                    default=prop.default_value,
                    prop=prop,
                    pydantic_field=pydantic_field,
                )
        elif isinstance(prop, dm.MappedProperty) and isinstance(prop.type, dm.DirectRelation):
            # For direct relation the source is required.
            view_id = cast(dm.ViewId, prop.source)
            target_data_class = data_class_by_view_id[ViewSpaceExternalId(view_id.space, view_id.external_id)]
            return EdgeOneToOne(
                name=name,
                prop_name=prop_name,
                prop=prop,
                data_class=target_data_class,
                pydantic_field=pydantic_field,
            )

        else:
            raise NotImplementedError(f"Property type={type(prop)!r} is not supported")

    @abstractmethod
    def as_read_type_hint(self) -> str:
        raise NotImplementedError()

    @abstractmethod
    def as_write_type_hint(self) -> str:
        raise NotImplementedError()

    @property
    @abstractmethod
    def is_edge(self) -> bool:
        raise NotImplementedError()

    @property
    @abstractmethod
    def is_time_field(self) -> bool:
        raise NotImplementedError()


@dataclass(frozen=True)
class PrimitiveFieldCore(Field, ABC):
    type_: str
    is_nullable: bool
    prop: dm.MappedProperty

    @property
    def is_edge(self) -> bool:
        return False

    @property
    def is_time_field(self) -> bool:
        return self.type_ in ("datetime.datetime", "datetime.date")


@dataclass(frozen=True)
class PrimitiveField(PrimitiveFieldCore):
    """
    This represents a basic type such as str, int, float, bool, datetime.datetime, datetime.date.
    """

    default: str | int | dict | None = None

    def as_read_type_hint(self) -> str:
        rhs = str(self.default)
        if self.need_alias:
            rhs = f'{self.pydantic_field}({rhs}, alias="{self.prop_name}")'

        return f"Optional[{self.type_}] = {rhs}"

    def as_write_type_hint(self) -> str:
        out_type = self.type_
        if self.is_nullable:
            out_type = f"Optional[{out_type}] = {self.default}"
        return out_type


@dataclass(frozen=True)
class PrimitiveListField(PrimitiveFieldCore):
    """
    This represents a list of basic types such as list[str], list[int], list[float], list[bool],
    list[datetime.datetime], list[datetime.date].
    """

    def as_read_type_hint(self) -> str:
        alias = ""
        if self.need_alias:
            alias = f', alias="{self.prop_name}"'

        if alias:
            rhs = f" = {self.pydantic_field}(default_factory=list{alias})"
        else:
            rhs = " = []"
        return f"list[{self.type_}]{rhs}"

    def as_write_type_hint(self) -> str:
        rhs = ""
        if self.is_nullable:
            rhs = " = []"
        return f"list[{self.type_}]{rhs}"


@dataclass(frozen=True)
class CDFExternalField(PrimitiveFieldCore):
    edge_api_class: str
    edge_api_attribute: str

    def as_read_type_hint(self) -> str:
        rhs = "None"
        if self.need_alias:
            rhs = f'{self.pydantic_field}({rhs}, alias="{self.prop_name}")'

        return f"Optional[{self.type_}] = {rhs}"

    def as_write_type_hint(self) -> str:
        out_type = self.type_
        if self.is_nullable:
            out_type = f"Optional[{out_type}] = None"
        return out_type


@dataclass(frozen=True)
class EdgeField(Field, ABC):
    """
    This represents an edge field linking to another data class.
    """

    data_class: DataClass

    @property
    def is_edge(self) -> bool:
        return True

    @property
    def is_time_field(self) -> bool:
        return False


@dataclass(frozen=True)
class EdgeOneToOne(EdgeField):
    """
    This represents an edge field linking to another data class.
    """

    prop: dm.MappedProperty

    def as_read_type_hint(self) -> str:
        return "Optional[str] = None"

    def as_write_type_hint(self) -> str:
        return f"Union[{self.data_class.write_name}, str, None] = {self.pydantic_field}(None, repr=False)"


@dataclass(frozen=True)
class EdgeOneToMany(EdgeField):
    """
    This represents a list of edge fields linking to another data class.
    """

    variable: str
    edge_api_class: str
    edge_api_attribute: str
    prop: dm.SingleHopConnectionDefinition

    def as_read_type_hint(self) -> str:
        return "list[str] = []"

    def as_write_type_hint(self) -> str:
        return (
            f"Union[list[{self.data_class.write_name}], list[str]] "
            f"= {self.pydantic_field}(default_factory=list, repr=False)"
        )


@dataclass(frozen=True)
class DataClass:
    """
    This represents a data class. It is created from a view.
    """

    view_name: str
    read_name: str
    write_name: str
    read_list_name: str
    write_list_name: str
    variable: str
    variable_list: str
    file_name: str
    view_id: ViewSpaceExternalId
    fields: list[Field] = field(default_factory=list)

    @classmethod
    def from_view(cls, view: dm.View, data_class: pygen_config.DataClassNaming) -> Self:
        view_name = (view.name or view.external_id).replace(" ", "_")
        class_name = create_name(view_name, data_class.name)
        variable_name = create_name(view_name, data_class.variable)
        variable_list = create_name(view_name, data_class.variable_list)
        if variable_name == variable_list:
            variable_list = f"{variable_list}_list"
        file_name = f"_{create_name(view_name, data_class.file)}"
        return cls(
            view_name=view_name,
            read_name=class_name,
            write_name=f"{class_name}Apply",
            read_list_name=f"{class_name}List",
            write_list_name=f"{class_name}ApplyList",
            variable=variable_name,
            variable_list=variable_list,
            file_name=file_name,
            view_id=ViewSpaceExternalId.from_(view),
        )

    def update_fields(
        self,
        properties: dict[str, dm.MappedProperty | dm.ConnectionDefinition],
        data_class_by_view_id: dict[ViewSpaceExternalId, DataClass],
        field_naming: pygen_config.FieldNaming,
    ) -> None:
        pydantic_field = self.pydantic_field
        for prop_name, prop in properties.items():
            field_ = Field.from_property(
                prop_name, prop, data_class_by_view_id, field_naming, self.view_name, pydantic_field=pydantic_field
            )
            self.fields.append(field_)

    @property
    def pydantic_field(self) -> str:
        if any(
            name == "Field" for name in [self.read_name, self.write_name, self.read_list_name, self.write_list_name]
        ):
            return "pydantic.Field"
        else:
            return "Field"

    @property
    def init_import(self) -> str:
        return (
            f"from .{self.file_name} "
            f"import {self.read_name}, {self.write_name}, {self.read_list_name}, {self.write_list_name}"
        )

    def __iter__(self) -> Iterator[Field]:
        return iter(self.fields)

    @property
    def one_to_one_edges(self) -> Iterable[EdgeOneToOne]:
        return (field_ for field_ in self.fields if isinstance(field_, EdgeOneToOne))

    @property
    def one_to_many_edges(self) -> Iterable[EdgeOneToMany]:
        return (field_ for field_ in self.fields if isinstance(field_, EdgeOneToMany))

    @property
    def primitive_fields(self) -> Iterable[PrimitiveField]:
        return (field_ for field_ in self.fields if isinstance(field_, PrimitiveField))

    @property
    def cdf_external_fields(self) -> Iterable[CDFExternalField]:
        return (field_ for field_ in self.fields if isinstance(field_, CDFExternalField))

    @property
    def single_timeseries_fields(self) -> Iterable[CDFExternalField]:
        return (field_ for field_ in self.cdf_external_fields if isinstance(field_.prop.type, dm.TimeSeriesReference))

    @property
    def has_one_to_many_edges(self) -> bool:
        return any(isinstance(field_, EdgeOneToMany) for field_ in self.fields)

    @property
    def has_edges(self) -> bool:
        return any(isinstance(field_, EdgeField) for field_ in self.fields)

    @property
    def has_only_one_to_many_edges(self) -> bool:
        return all(isinstance(field_, EdgeOneToMany) for field_ in self.fields)

    @property
    def fields_by_container(self) -> dict[dm.ContainerId, list[PrimitiveFieldCore | EdgeOneToOne]]:
        result: dict[dm.ContainerId, list[PrimitiveFieldCore | EdgeOneToOne]] = defaultdict(list)
        for field_ in self:
            if isinstance(field_, (PrimitiveFieldCore, EdgeOneToOne)):
                result[field_.prop.container].append(field_)
        return dict(result)

    @property
    def has_time_field(self) -> bool:
        return any(field_.is_time_field for field_ in self.fields)

    @property
    def _field_type_hints(self) -> Iterable[str]:
        return (hint for field_ in self.fields for hint in (field_.as_read_type_hint(), field_.as_write_type_hint()))

    @property
    def use_optional_type(self) -> bool:
        return any("Optional" in hint for hint in self._field_type_hints)

    @property
    def use_pydantic_field(self) -> bool:
        pydantic_field = self.pydantic_field
        return any(pydantic_field in hint for hint in self._field_type_hints)

    @property
    def import_pydantic_field(self) -> str:
        if self.pydantic_field == "Field":
            return "from pydantic import Field"
        else:
            return "import pydantic"

    @property
    def dependencies(self) -> list[DataClass]:
        unique: dict[ViewSpaceExternalId, DataClass] = {}
        for field_ in self.fields:
            if isinstance(field_, EdgeField):
                # This will overwrite any existing data class with the same view id
                # however, this is not a problem as all data classes are uniquely identified by their view id
                unique[field_.data_class.view_id] = field_.data_class
        return sorted(unique.values(), key=lambda x: x.write_name)

    @property
    def has_single_timeseries_fields(self) -> bool:
        return any(
            isinstance(field_.prop.type, dm.TimeSeriesReference) and not isinstance(field_, PrimitiveListField)
            for field_ in self.single_timeseries_fields
        )

    @property
    def primitive_fields_literal(self) -> str:
        return ", ".join(
            f'"{field_.prop_name}"' for field_ in self if isinstance(field_, (PrimitiveField, CDFExternalField))
        )


@dataclass(frozen=True)
class APIClass:
    client_attribute: str
    name: str
    file_name: str
    view_id: ViewSpaceExternalId

    @classmethod
    def from_view(cls, view: dm.View, api_class: pygen_config.APIClassNaming) -> APIClass:
        raw_name = view.name or view.external_id

        raw_name = raw_name.replace(" ", "_")

        return cls(
            client_attribute=create_name(raw_name, api_class.client_attribute),
            name=f"{create_name(raw_name, api_class.name)}API",
            file_name=create_name(raw_name, api_class.file_name),
            view_id=ViewSpaceExternalId.from_(view),
        )


@dataclass(frozen=True)
class MultiAPIClass:
    """
    This represents a set of APIs which are generated from a single data model.

    The motivation for having this class is the case when you want to create one SDK for multiple data models.
    """

    sub_apis: list[APIClass]
    client_attribute: str
    name: str
    model: dm.DataModel[dm.View]

    @property
    def model_id(self) -> dm.DataModelId:
        return self.model.as_id()

    @classmethod
    def from_data_model(
        cls,
        data_model: dm.DataModel[dm.View],
        api_class_by_view_id: dict[ViewSpaceExternalId, APIClass],
        multi_api_class: pygen_config.MultiAPIClassNaming,
    ) -> MultiAPIClass:
        sub_apis = sorted(
            [api_class_by_view_id[ViewSpaceExternalId.from_(view)] for view in data_model.views],
            key=lambda api: api.name,
        )

        data_model_name = data_model.name or data_model.external_id

        return cls(
            sub_apis=sub_apis,
            client_attribute=create_name(data_model_name, multi_api_class.client_attribute),
            name=f"{create_name(data_model_name, multi_api_class.name)}APIs",
            model=data_model,
        )


@dataclass
class ListParameter:
    name: str
    type_: str
    default: None = None

    @property
    def annotation(self) -> str:
        return f"{self.type_} | None"

    @property
    def is_time(self) -> bool:
        return self.type_ in ("datetime.datetime", "datetime.date")


@dataclass
class ListFilter:
    filter: type[dm.Filter]
    prop_name: str
    keyword_arguments: dict[str, ListParameter]

    @property
    def condition(self) -> str:
        if self.filter is dm.filters.In:
            parameter = next(iter(self.keyword_arguments.values())).name
            return f"{parameter} and isinstance({parameter}, list)"
        elif self.filter is dm.filters.Equals:
            parameter = next(iter(self.keyword_arguments.values())).name
            return f"{parameter} and isinstance({parameter}, str)"

        return " or ".join(arg.name for arg in self.keyword_arguments.values())

    @property
    def arguments(self) -> str:
        if self.prop_name == "externalId":
            property_ref = '["node", "externalId"], '
        else:
            property_ref = f'view_id.as_property_ref("{self.prop_name}"), '

        filter_args = ", ".join(
            (
                f"{keyword}={arg.name}.isoformat() if {arg.name} else None" if arg.is_time else f"{keyword}={arg.name}"
                for keyword, arg in self.keyword_arguments.items()
            )
        )
        return f"{property_ref}{filter_args}"

    @property
    def filter_call(self) -> str:
        return f"dm.filters.{self.filter.__name__}"


# This field is used when creating the list method.
_EXTERNAL_ID_FIELD = PrimitiveField(
    name="external_id",
    prop_name="externalId",
    type_="str",
    is_nullable=False,
    default=None,
    prop=dm.MappedProperty(
        container_property_identifier="externalId",
        type=dm.Text(),
        nullable=False,
        auto_increment=False,
        container=dm.ContainerId("dummy", "dummy"),
    ),
    pydantic_field="Field",
)


@dataclass
class ListMethod:
    parameters: list[ListParameter]
    filters: list[ListFilter]

    @classmethod
    def from_fields(cls, fields: Iterable[Field], config: pygen_config.ListMethodFilters) -> Self:
        parameters_by_name: dict[str, ListParameter] = {}
        list_filters: list[ListFilter] = []

        for field_ in itertools.chain(fields, (_EXTERNAL_ID_FIELD,)):
            if not isinstance(field_, PrimitiveField):
                # Only primitive fields supported for now
                continue

            for selected_filter in config.get(field_.prop.type, field_.prop_name):
                if selected_filter is dm.filters.Equals:
                    if field_.name not in parameters_by_name:
                        parameter = ListParameter(name=field_.name, type_=field_.type_)
                        parameters_by_name[parameter.name] = parameter
                    else:
                        # Equals and In filter share parameter, you have to extend the type hint.
                        parameter = parameters_by_name[field_.name]
                        parameter.type_ = f"{field_.type_} | {parameter.type_}"
                    list_filters.append(
                        ListFilter(
                            filter=selected_filter,
                            prop_name=field_.prop_name,
                            keyword_arguments=dict(value=parameter),
                        )
                    )
                elif selected_filter is dm.filters.In:
                    if field_.name not in parameters_by_name:
                        parameter = ListParameter(field_.name, type_=f"list[{field_.type_}]")
                        parameters_by_name[parameter.name] = parameter
                    else:
                        # Equals and In filter share parameter, you have to extend the typ hint.
                        parameter = parameters_by_name[field_.name]
                        parameter.type_ = f"{parameter.type_} | list[{field_.type_}]"
                    list_filters.append(
                        ListFilter(
                            filter=selected_filter,
                            prop_name=field_.prop_name,
                            keyword_arguments=dict(values=parameter),
                        )
                    )
                elif selected_filter is dm.filters.Prefix:
                    parameter = ListParameter(name=f"{field_.name}_prefix", type_=field_.type_)
                    parameters_by_name[parameter.name] = parameter
                    list_filters.append(
                        ListFilter(
                            filter=selected_filter,
                            prop_name=field_.prop_name,
                            keyword_arguments=dict(value=parameter),
                        )
                    )
                elif selected_filter is dm.filters.Range:
                    min_parameter = ListParameter(name=f"min_{field_.name}", type_=field_.type_)
                    max_parameter = ListParameter(name=f"max_{field_.name}", type_=field_.type_)
                    parameters_by_name[min_parameter.name] = min_parameter
                    parameters_by_name[max_parameter.name] = max_parameter
                    list_filters.append(
                        ListFilter(
                            filter=selected_filter,
                            prop_name=field_.prop_name,
                            keyword_arguments=dict(gte=min_parameter, lte=max_parameter),
                        )
                    )
                else:
                    # This is a filter that is not supported by the list method.
                    continue
        return cls(
            parameters=list(parameters_by_name.values()),
            filters=list_filters,
        )


def _to_python_type(type_: dm.DirectRelationReference | dm.PropertyType) -> str:
    if isinstance(type_, (dm.Int32, dm.Int64)):
        out_type = "int"
    elif isinstance(type_, dm.Boolean):
        out_type = "bool"
    elif isinstance(type_, (dm.Float32, dm.Float64)):
        out_type = "float"
    elif isinstance(type_, dm.Date):
        out_type = "datetime.date"
    elif isinstance(type_, dm.Timestamp):
        out_type = "datetime.datetime"
    elif isinstance(type_, dm.Json):
        out_type = "dict"
    elif isinstance(type_, (dm.Text, dm.DirectRelation, dm.CDFExternalIdReference, dm.DirectRelationReference)):
        out_type = "str"
    else:
        raise ValueError(f"Unknown type {type_}")

    return out_type



from __future__ import annotations

from collections.abc import Sequence
from pathlib import Path
from typing import Callable, Literal

from cognite.client import data_modeling as dm
from cognite.client._version import __version__ as cognite_sdk_version
from jinja2 import Environment, PackageLoader, select_autoescape
from pydantic.version import VERSION as PYDANTIC_VERSION

from cognite.pygen._version import __version__
from cognite.pygen.config import PygenConfig
from cognite.pygen.utils.helper import get_pydantic_version

from . import validation
from .data_classes import APIClass, DataClass, ListMethod, MultiAPIClass, ViewSpaceExternalId
from .logic import get_unique_views


class SDKGenerator:
    """
    SDK generator for a data model.

    Args:
        top_level_package: The name of the top level package for the SDK. Example "movie.client"
        client_name: The name of the client class. Example "MovieClient"
        data_model: The data model(s) to generate a SDK for.
        pydantic_version: The version of pydantic to use. "infer" will use the version of pydantic installed in
                          the environment.
        logger: A logger function to use for logging. If None, print will be done.
    """

    def __init__(
        self,
        top_level_package: str,
        client_name: str,
        data_model: dm.DataModel | Sequence[dm.DataModel],
        pydantic_version: Literal["v1", "v2", "infer"] = "infer",
        logger: Callable[[str], None] | None = None,
        config: PygenConfig = PygenConfig(),
    ):
        self._data_model = data_model
        self.top_level_package = top_level_package
        self.client_name = client_name
        self._multi_api_classes: list[MultiAPIClass]
        if isinstance(data_model, dm.DataModel):
            self._multi_api_generator = MultiAPIGenerator(
                top_level_package, client_name, data_model.views, pydantic_version, logger, config
            )
            self._multi_api_classes = []
        elif isinstance(data_model, Sequence):
            unique_views = get_unique_views(*[view for model in data_model for view in model.views])

            self._multi_api_generator = MultiAPIGenerator(
                top_level_package, client_name, unique_views, pydantic_version, logger, config
            )
            api_by_view_identifier = {api.view_identifier: api.api_class for api in self._multi_api_generator.sub_apis}

            self._multi_api_classes = sorted(
                (
                    MultiAPIClass.from_data_model(model, api_by_view_identifier, config.naming.multi_api_class)
                    for model in data_model
                ),
                key=lambda a: a.name,
            )
            validation.validate_multi_api_classes(self._multi_api_classes)
        else:
            raise ValueError("data_model must be a DataModel or a sequence of DataModels")

    def generate_sdk(self) -> dict[Path, str]:
        """
        Generate the SDK.

        Returns:
            A Python SDK given as a dictionary of file paths and file contents, which can be written to disk.
        """
        client_dir = Path(self.top_level_package.replace(".", "/"))
        sdk = self._multi_api_generator.generate_apis(client_dir)
        sdk[client_dir / "_api_client.py"] = self._generate_api_client_file()
        return sdk

    def _generate_api_client_file(self) -> str:
        api_client = self._multi_api_generator.env.get_template("_api_client.py.jinja")

        api_classes = sorted(
            (sub_api.api_class for sub_api in self._multi_api_generator.sub_apis),
            key=lambda api_class: api_class.name.lower(),
        )

        # In the template, we run zip(api_classes, views) and zip(multi_api_classes, view_sets)
        # thus it is important that the order is the same for both.
        views, view_sets = self._create_view_lists_in_order_of_api_classes(api_classes)

        return (
            api_client.render(
                client_name=self.client_name,
                pygen_version=__version__,
                cognite_sdk_version=cognite_sdk_version,
                pydantic_version=PYDANTIC_VERSION,
                api_classes=api_classes,
                is_single_model=isinstance(self._data_model, dm.DataModel),
                top_level_package=self.top_level_package,
                data_model=self._data_model,
                multi_api_classes=self._multi_api_classes,
                view_sets=view_sets,
                views=views,
                zip=zip,
            )
            + "\n"
        )

    def _create_view_lists_in_order_of_api_classes(
        self, api_classes: list[APIClass]
    ) -> tuple[list[dm.View], list[list[dm.View]]]:
        api_class_order: dict[ViewSpaceExternalId, int] = {
            api_class.view_id: i for i, api_class in enumerate(api_classes)
        }

        if isinstance(self._data_model, dm.DataModel):
            views = sorted(
                self._data_model.views,
                key=lambda v: api_class_order[ViewSpaceExternalId.from_(v)],
            )
            return views, []
        elif isinstance(self._data_model, Sequence):
            multi_api_class_order = {api_class.model_id: i for i, api_class in enumerate(self._multi_api_classes)}
            sorted_data_models = sorted(  # type: ignore[arg-type]
                self._data_model,
                key=lambda m: multi_api_class_order[m.as_id()],
            )
            view_sets: list[list[dm.View]] = []
            for data_model in sorted_data_models:
                views = sorted(data_model.views, key=lambda v: api_class_order[ViewSpaceExternalId.from_(v)])
                view_sets.append(views)
            return [], view_sets
        else:
            raise ValueError("data_model must be a DataModel or a sequence of DataModels")


class MultiAPIGenerator:
    def __init__(
        self,
        top_level_package: str,
        client_name: str,
        views: Sequence[dm.View],
        pydantic_version: Literal["v1", "v2", "infer"] = "infer",
        logger: Callable[[str], None] | None = None,
        config: PygenConfig = PygenConfig(),
    ):
        self.env = Environment(loader=PackageLoader("cognite.pygen._core", "templates"), autoescape=select_autoescape())
        self.top_level_package = top_level_package
        self.client_name = client_name
        self._pydantic_version = pydantic_version
        self._logger = logger or print

        self.sub_apis: list[APIGenerator] = [APIGenerator(view, config) for view in views]
        data_class_by_view_id: dict[ViewSpaceExternalId, DataClass] = {
            api.view_identifier: api.data_class for api in self.sub_apis
        }
        for api, view in zip(self.sub_apis, views):
            api.data_class.update_fields(view.properties, data_class_by_view_id, config.naming.field)

        validation.validate_data_classes([api.data_class for api in self.sub_apis])
        validation.validate_api_classes([api.api_class for api in self.sub_apis])

    def __getitem__(self, item: dm.View | dm.ViewId) -> APIGenerator | None:
        return next(
            (api for api in self.sub_apis if api.view_identifier == ViewSpaceExternalId.from_(item)),
            None,
        )

    @property
    def pydantic_version(self) -> Literal["v1", "v2"]:
        if self._pydantic_version == "infer":
            return get_pydantic_version()
        elif self._pydantic_version in ["v1", "v2"]:
            return self._pydantic_version
        else:
            raise ValueError(f"Unknown pydantic version {self._pydantic_version}")

    def generate_apis(self, client_dir: Path) -> dict[Path, str]:
        data_classes_dir = client_dir / "data_classes"
        api_dir = client_dir / "_api"

        sdk = {(api_dir / "__init__.py"): ""}
        for api in self.sub_apis:
            file_name = api.api_class.file_name
            sdk[data_classes_dir / f"_{file_name}.py"] = api.generate_data_class_file()
            sdk[api_dir / f"{file_name}.py"] = api.generate_api_file(self.top_level_package)

        sdk[client_dir / "__init__.py"] = self.generate_client_init_file()
        sdk[data_classes_dir / "__init__.py"] = self.generate_data_classes_init_file()
        sdk[api_dir / "_core.py"] = self.generate_api_core_file()

        sdk[data_classes_dir / "_core.py"] = self.generate_data_class_core_file()
        # if self.pydantic_version == "v2":
        return sdk

    def generate_api_core_file(self) -> str:
        api_core = self.env.get_template("_core_api.py.jinja")
        return api_core.render(top_level_package=self.top_level_package) + "\n"

    def generate_data_class_core_file(self) -> str:
        data_class_core = self.env.get_template("_core_data_classes.py.jinja")
        return data_class_core.render(is_pydantic_v2=self.pydantic_version == "v2") + "\n"

    def generate_client_init_file(self) -> str:
        client_init = self.env.get_template("_client_init.py.jinja")
        return client_init.render(client_name=self.client_name, top_level_package=self.top_level_package) + "\n"

    def generate_data_classes_init_file(self) -> str:
        data_class_init = self.env.get_template("data_classes_init.py.jinja")

        data_classes_with_dependencies = sorted(
            (api.data_class for api in self.sub_apis if api.data_class.has_edges), key=lambda d: d.read_name
        )
        dependencies_by_data_class_write_name = {
            data_class.write_name: sorted(data_class.dependencies, key=lambda d: d.read_name)
            for data_class in data_classes_with_dependencies
        }

        return (
            data_class_init.render(
                classes=sorted((api.data_class for api in self.sub_apis), key=lambda d: d.read_name),
                dependencies_by_data_class_write_name=dependencies_by_data_class_write_name,
                top_level_package=self.top_level_package,
                import_file={
                    "v2": "data_classes_init_import.py.jinja",
                    "v1": "data_classes_init_import.py_pydanticv1.jinja",
                }[self.pydantic_version],
            )
            + "\n"
        )


class APIGenerator:
    def __init__(self, view: dm.View, config: PygenConfig):
        self._env = Environment(
            loader=PackageLoader("cognite.pygen._core", "templates"), autoescape=select_autoescape()
        )
        self.view_identifier = ViewSpaceExternalId.from_(view)
        self.data_class = DataClass.from_view(view, config.naming.data_class)
        self.api_class = APIClass.from_view(view, config.naming.api_class)

        # List method cannot be generated here, as we need all data class fields to bo updated first.
        self._config = config

    def generate_data_class_file(self) -> str:
        type_data = self._env.get_template("data_class.py.jinja")

        return type_data.render(data_class=self.data_class, space=self.view_identifier.space) + "\n"

    def generate_api_file(self, top_level_package: str) -> str:
        type_api = self._env.get_template("api_class.py.jinja")

        return (
            type_api.render(
                top_level_package=top_level_package,
                api_class=self.api_class,
                data_class=self.data_class,
                list_method=ListMethod.from_fields(self.data_class.fields, self._config.list_method),
            )
            + "\n"
        )



from __future__ import annotations

import hashlib
import json
from typing import Any

from cognite.client.data_classes import data_modeling as dm


def _unique_properties(
    prop: dm.MappedProperty | dm.SingleHopConnectionDefinition | dm.MappedProperty | dm.ConnectionDefinition,
) -> dict[str, Any]:
    if isinstance(prop, dm.MappedProperty):
        return {
            "container": prop.container.dump(),
            "container_property_identifier": prop.container_property_identifier,
            "default_value": prop.default_value,
            "name": prop.name,
            "nullable": prop.nullable,
            "type": prop.type.dump(),
        }
    elif isinstance(prop, dm.SingleHopConnectionDefinition):
        return {
            "direction": prop.direction,
            "name": prop.name,
            "type": prop.type.dump(),
        }
    else:
        raise ValueError(f"Unknown property type {prop}")


def _unique_views_properties(view: dm.View) -> dict[str, Any]:
    """
    Returns the properties from a view that uniquely defines it.

    This is necessary as there might be two views that have different versions, but all else is the same,
    thus they can be used to create the same data classes and APIs in the generated SDK. In other words,
    a data class and API class can be reused for two views if they have the same properties.

    Args:
        view: The View

    Returns
        A dictionary with the properties that uniquely defines the view.
    """
    return {
        "space": view.space,
        "name": view.name,
        "externalId": view.external_id,
        "properties": {name: _unique_properties(prop) for name, prop in view.properties.items()},
    }


def get_unique_views(*views: dm.View) -> list[dm.View]:
    view_hashes = set()
    unique_views = []
    for view in views:
        view_hash = hashlib.shake_256(json.dumps(_unique_views_properties(view)).encode("utf-8")).hexdigest(16)
        if view_hash not in view_hashes:
            unique_views.append(view)
            view_hashes.add(view_hash)
    return unique_views



from pathlib import Path
from typing import Annotated, cast

from cognite.client import CogniteClient
from cognite.client.exceptions import CogniteAPIError

from cognite import pygen
from cognite.pygen import generate_sdk
from cognite.pygen._settings import PygenSettings, load_settings

try:
    import typer
except ImportError:
    _has_typer = False
else:
    _has_typer = True


def _version_callback(value: bool):
    if value:
        typer.echo(pygen.__version__)
        raise typer.Exit()


if _has_typer:
    app = typer.Typer(add_completion=False)

    @app.callback()
    def common(
        ctx: typer.Context,
        version: bool = typer.Option(None, "--version", callback=_version_callback),
    ):
        ...

    settings = load_settings()
    help_text = "Generate a Python SDK from Data Model(s)"
    if settings is not None:
        loaded_settings = cast(PygenSettings, settings)

        @app.command(help=help_text)
        def generate(
            space: str = typer.Option(default=loaded_settings.space.default, help=loaded_settings.space.help),
            external_id: str = typer.Option(
                default=loaded_settings.external_id.default, help=loaded_settings.external_id.help
            ),
            version: str = typer.Option(default=loaded_settings.version.default, help=loaded_settings.version.help),
            tenant_id: str = typer.Option(
                default=loaded_settings.tenant_id.default, help=loaded_settings.tenant_id.help
            ),
            client_id: str = typer.Option(
                default=loaded_settings.client_id.default, help=loaded_settings.client_id.help
            ),
            client_secret: str = typer.Option(
                default=loaded_settings.client_secret.default,
                help=loaded_settings.client_secret.help,
                show_default=False,
            ),
            cdf_cluster: str = typer.Option(
                default=loaded_settings.cdf_cluster.default, help=loaded_settings.cdf_cluster.help
            ),
            cdf_project: str = typer.Option(
                default=loaded_settings.cdf_project.default, help=loaded_settings.cdf_project.help
            ),
            output_dir: Path = typer.Option(
                default=loaded_settings.output_dir.default or Path.cwd(), help=loaded_settings.output_dir.help
            ),
            top_level_package: str = typer.Option(
                loaded_settings.top_level_package.default, help=loaded_settings.top_level_package.help
            ),
            client_name: str = typer.Option(loaded_settings.client_name.default, help=loaded_settings.client_name.help),
            overwrite: bool = typer.Option(loaded_settings.overwrite.default, help=loaded_settings.overwrite.help),
            skip_formatting: bool = typer.Option(
                loaded_settings.skip_formatting.default, help=loaded_settings.skip_formatting.help
            ),
        ):
            client = CogniteClient.default_oauth_client_credentials(
                cdf_project, cdf_cluster, tenant_id, client_id, client_secret
            )
            data_models: list[tuple[str, str, str]]
            if settings and settings.data_models:
                data_models = settings.data_models
            else:
                data_models = []
            if space is not None and external_id is not None and version is not None:
                data_models = [(space, external_id, version)]
            if not data_models:
                raise ValueError(
                    "No data model(s) specified. Please either with --space --external-id --version "
                    "or 'data_models' in the 'pyproject.toml' file."
                )
            try:
                generate_sdk(
                    client,
                    data_models,
                    top_level_package,
                    client_name,
                    output_dir,
                    typer.echo,
                    overwrite=overwrite,
                    format_code=not skip_formatting,
                )
            except (CogniteAPIError, IndexError) as e:
                raise typer.Exit(code=1) from e

    else:
        default_settings = PygenSettings()

        @app.command(help=help_text)
        def generate(
            space: Annotated[str, typer.Option(..., help=default_settings.space.help)],
            external_id: Annotated[str, typer.Option(..., help=default_settings.external_id.help)],
            version: Annotated[str, typer.Option(..., help=default_settings.version.help)],
            tenant_id: Annotated[str, typer.Option(..., help=default_settings.tenant_id.help)],
            client_id: Annotated[str, typer.Option(..., help=default_settings.client_id.help)],
            client_secret: Annotated[str, typer.Option(..., help="Azure Client Secret for connecting to CDF")],
            cdf_cluster: Annotated[str, typer.Option(..., help=default_settings.cdf_cluster.help)],
            cdf_project: Annotated[str, typer.Option(..., help=default_settings.cdf_project.help)],
            output_dir: Path = typer.Option(Path.cwd(), help=default_settings.output_dir.help),
            top_level_package: str = typer.Option(
                default_settings.top_level_package.default, help=default_settings.top_level_package.help
            ),
            client_name: str = typer.Option(
                default_settings.client_name.default, help=default_settings.client_name.help
            ),
            overwrite: bool = typer.Option(default_settings.overwrite.default, help=default_settings.overwrite.help),
            skip_formatting: bool = typer.Option(
                default_settings.skip_formatting.default, help=default_settings.skip_formatting.help
            ),
        ):
            client = CogniteClient.default_oauth_client_credentials(
                cdf_project, cdf_cluster, tenant_id, client_id, client_secret
            )
            try:
                generate_sdk(
                    client,
                    (space, external_id, version),
                    top_level_package,
                    client_name,
                    output_dir,
                    typer.echo,
                    overwrite=overwrite,
                    format_code=not skip_formatting,
                )
            except (CogniteAPIError, IndexError) as e:
                raise typer.Exit(code=1) from e

    def main():
        app()

else:

    def main():
        print("THE CLI requires typer to be available, install with `pip install cognite-pygen[cli]`")


if __name__ == "__main__":
    main()



from ._core.generators import SDKGenerator
from ._generator import generate_sdk, generate_sdk_notebook, write_sdk_to_disk
from ._version import __version__
from .utils.cdf import load_cognite_client_from_toml

__all__ = [
    "__version__",
    "load_cognite_client_from_toml",
    "generate_sdk",
    "generate_sdk_notebook",
    "write_sdk_to_disk",
    "SDKGenerator",
]
